import datetime
import random
from pathlib import Path
from typing import Any, Protocol

import peewee

DEFAULT_DB_NAME = "oeleo-file-list.db"

CODES = [
    (0, "should-be-copied"),
    (1, "should-be-copied-if-changed"),
    (2, "should-not-be-copied"),
]

database_proxy = peewee.DatabaseProxy()


class FileList(peewee.Model):
    processed_date = peewee.DateTimeField(default=datetime.datetime.now())
    local_name = peewee.CharField(unique=True)
    external_name = peewee.CharField(null=True)
    checksum = peewee.CharField(null=True)
    code = peewee.SmallIntegerField(choices=CODES, default=0)

    class Meta:
        database = database_proxy


class DbHandler(Protocol):
    def initialize_db(self):
        ...

    def register(self, f: Path):
        ...

    def is_changed(self, **checks: Any) -> bool:
        ...

    def update_record(self, external_name: Path, code: int = 1, **checks: Any):
        ...


class MockDbHandler(DbHandler):
    def __init__(self):
        self.db_name = "mock"

    def initialize_db(self):
        print("INITIALIZING DB")

    def register(self, f: Path):
        print(f"REGISTERING {f}")

    def is_changed(self, **checks) -> bool:
        print(f"CHECKING IF IT HAS CHANGED")
        print(f"checks: {checks}")
        _is_changed = random.choice([True, False])
        print(f"is-changed: {_is_changed}")
        return _is_changed

    def update_record(self, external_name: Path, code: int = 1, **checks: Any):
        print("UPDATE RECORD IN DB")
        print(f"external name: {external_name}")
        print(f"additional checks: {checks}")


class SimpleDbHandler(DbHandler):
    """A simple db bookkeeper using sqlite3 that checks on checksum."""

    def __init__(self, db_name: str) -> None:
        self.db_name = db_name
        self.db_model: peewee.Model = FileList
        self._current_record = None
        self.db_instance = database_proxy
        self._set_up_sqlite_db()

    @staticmethod
    def _set_up_sqlite_db():
        database_proxy.initialize(peewee.SqliteDatabase(None))

    @property
    def record(self):
        return self._current_record

    @record.setter
    def record(self, value):
        self._current_record = value

    @record.deleter
    def record(self):
        self._current_record = None

    def initialize_db(self):
        self.db_instance.init(self.db_name)
        self.db_instance.connect()
        self.db_instance.create_tables([self.db_model])

    def register(self, f: Path):
        """Get or create record of the file and check if it needs to be copied."""
        self._current_record, new = self.db_model.get_or_create(local_name=f.name)
        if new:
            self.record.code = 0
            self.record.save()

    def is_changed(self, **checks) -> bool:
        if self.record.code == 0:
            return True

        _is_changed = False
        for k in checks:
            try:
                v = getattr(self.record, k)
            except AttributeError as e:
                raise AttributeError("oeleo-model-key mismatch") from e
            if v != checks[k]:
                _is_changed = True
        return _is_changed

    def update_record(self, external_name: Path, code: int = 1, **checks: Any):
        checksum = checks.get("checksum", None)
        self.record.checksum = checksum
        self.record.processed_date = datetime.datetime.now()
        self.record.code = code
        self.record.external_name = external_name
        self.record.save()

    @property
    def code(self):
        return self.record.code

    @code.setter
    def code(self, code):
        if code not in (0, 1, 2):
            raise ValueError(f"code is not valid")
        self.record.code = code
        self.record.save()



import hashlib
import logging
from pathlib import Path

import peewee
from rich.logging import RichHandler


def calculate_checksum(file_path: Path) -> str:
    with open(file_path, "rb") as f:
        file_hash = hashlib.md5()
        while chunk := f.read(8192):
            file_hash.update(chunk)
    return file_hash.hexdigest()


def logger(name="oeleo", log_level=logging.INFO, log_message_format=None):
    log_message_format = log_message_format or "%(message)s"

    logging.basicConfig(
        format=log_message_format,
        datefmt="[%X]",
        handlers=[RichHandler(rich_tracebacks=True, tracebacks_suppress=[peewee])],
    )

    log = logging.getLogger(name)
    log.setLevel(log_level)
    return log



import logging
from datetime import datetime
from functools import partial
from pathlib import Path
from typing import Any, Generator, Iterable

log = logging.getLogger(__name__)


def filter_on_not_before(path: Path, value: datetime):
    st = path.stat()
    sdt = datetime.fromtimestamp(st.st_mtime)
    if sdt >= value:
        return True
    else:
        return False


def filter_on_not_after(path: Path, value: datetime):
    st = path.stat()
    sdt = datetime.fromtimestamp(st.st_mtime)
    if sdt <= value:
        return True
    else:
        return False


FILTERS = {
    "not_before": filter_on_not_before,
    "not_after": filter_on_not_after,
}

FilterFunction = Any
FilterTuple = tuple[str, FilterFunction]


def base_filter(
    path: Path,
    extension: str = None,
    additional_filters: Iterable[FilterTuple] = None,
    base_filter_func: Any = None,
) -> Generator[Path, None, None]:

    """Simple directory content filter - cannot be used for ssh (yet)?"""

    if base_filter_func is None:
        base_filter_func = path.glob

    file_list = base_filter_func(f"*.{extension}")
    if additional_filters is not None:
        for filter_name, filter_val in additional_filters:
            filter_func = FILTERS[filter_name]
            file_list = filter(partial(filter_func, value=filter_val), file_list)
    return file_list


def main():
    directory = Path(r"/check/from")
    extension = "*"
    not_before = datetime(year=2022, month=7, day=1, hour=1, minute=0, second=0)
    not_after = datetime(year=2022, month=7, day=4, hour=1, minute=0, second=0)
    print("Starting...")

    my_filters = [
        ("not_before", not_before),
        ("not_after", not_after),
    ]

    g = base_filter(directory, extension, additional_filters=my_filters)
    for i in g:
        print(i)


if __name__ == "__main__":
    main()



import logging
import os
from datetime import datetime
from pathlib import Path

import dotenv

from oeleo.connectors import SSHConnector, register_password
from oeleo.utils import logger
from oeleo.workers import simple_worker, ssh_worker

log = logger()


def main():
    log.setLevel(logging.INFO)
    log.info(f"Starting oeleo!")
    dotenv.load_dotenv()
    filter_extension = os.environ["OELEO_FILTER_EXTENSION"]
    worker = simple_worker()
    worker.connect_to_db()
    worker.filter_local(filter_extension)
    worker.check(filter_extension)
    worker.run()


def example_check_with_ssh_connection():
    print(" example_check_with_ssh_connection ".center(80, "-"))
    log.setLevel(logging.INFO)
    log.info(f"Starting oeleo!")
    dotenv.load_dotenv()

    external_dir = "/home/jepe@ad.ife.no/Temp"
    filter_extension = "res"

    register_password(os.environ["OELEO_PASSWORD"])
    connector = SSHConnector(directory=external_dir)
    connector.connect()

    worker = ssh_worker(
        db_name=r"test_databases\test_ssh_to_odin.db",
        base_directory_from=Path(r"C:\scripting\processing_cellpy\raw"),
        connector=connector,
    )
    worker.connect_to_db()
    try:
        worker.check(filter_extension, update_db=True)
        worker.filter_local(filter_extension)
        worker.run()
    finally:
        connector.close()


def example_check_first_then_run():
    print(" example_check_first_then_run ".center(80, "-"))
    not_before = datetime(year=2021, month=3, day=1, hour=1, minute=0, second=0)
    not_after = datetime(year=2022, month=7, day=1, hour=1, minute=0, second=0)
    print("Starting...")

    my_filters = [
        ("not_before", not_before),
        ("not_after", not_after),
    ]

    log.setLevel(logging.INFO)
    log.info(f"Starting oeleo!")
    dotenv.load_dotenv()
    filter_extension = "res"
    worker = simple_worker(
        db_name=r"test_databases\another.db",
        base_directory_from=Path(r"C:\scripting\processing_cellpy\raw"),
        base_directory_to=Path(r"C:\scripting\trash"),
    )
    worker.connect_to_db()
    worker.filter_local(filter_extension)
    worker.check(filter_extension, additional_filters=my_filters)
    run_oeleo = input("\n Continue ([y]/n)? ") or "y"
    if run_oeleo.lower() in ["y", "yes"]:
        worker.run()


if __name__ == "__main__":
    main()
    example_check_first_then_run()
    example_check_with_ssh_connection()



import logging
import os
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Generator

from oeleo.checkers import ConnectedChecker, SimpleChecker
from oeleo.connectors import Connector, SSHConnector
from oeleo.filters import base_filter
from oeleo.models import DbHandler, MockDbHandler, SimpleDbHandler
from oeleo.movers import connected_mover, mock_mover, simple_mover

log = logging.getLogger("oeleo")


@dataclass
class Worker:
    """The worker class is responsible for orchestrating the transfers.

    A typical transfer consists of the following steps:
    1. Asking the bookkeeper to connect to its database.
    >>> worker.connect_to_db()
    2. Collect (and filter) the files in the local directory that are candidates for copying to
       the external directory (server).
    >>> worker.filter_local("*.res", additional_filters=my_filters)
    3. Calculate checksum for each file collected and check copy if they have changed.
    >>> worker.run()

    A worker needs to be initiated with the at least the following handlers:
        checker: Checker object
        mover_method: mover method
        from_dir: Path of local directory
        to_dir: Path of external directory
        bookkeeper: DbHandler to interact with the db

    Optinal handlers:
        external_connector:  Connector class for handling "difficult" connections (e.g. ssh)."""

    checker: Any
    mover_method: Any
    from_dir: Path
    to_dir: Path
    bookkeeper: DbHandler
    dry_run: bool = False
    local_connector: Any = None
    external_connector: Any = None
    filter_method: Any = base_filter
    file_names: Generator[Path, None, None] = field(init=False, default=None)
    _external_name: str = field(init=False, default="")
    _status: dict = field(init=False, default_factory=dict)

    def __post_init__(self):
        if self.dry_run:
            log.info("[bold red blink]DRY RUN[/]", extra={"markup": True})
            self.mover_method = mock_mover
            self.bookkeeper = MockDbHandler()

    def connect_to_db(self):
        self.bookkeeper.initialize_db()
        log.info(f"Connecting to db -> '{self.bookkeeper.db_name}' DONE")

    def filter_local(self, *args, **kwargs):
        """Selects the files that should be processed"""
        self.file_names = self._filter(
            None,
            self.from_dir,
            *args,
            **kwargs,
        )
        log.info("Filtering -> DONE")

    def _filter(
        self, connector: Connector | None, dir_path: Path, extension, *args, **kwargs
    ) -> Generator | list:
        """Selects the files that should be checked .

        Arguments:
            connector: if using a Connector object (can be None).
            dir_path: path to files.
            extension: file extension filter on.
            args: transferred untouched.
        Optional keyword arguments:
            base_filter_func: deprecated - will be extracted from the connector instead.
            additional_filters: list of tuples describing additional filter to use (only for local files)
        """
        base_filter_func = kwargs.pop("base_filter_func", None)
        if (base_filter_func is None) and (connector is not None):
            base_filter_func = connector.base_filter_sub_method
        file_names = self.filter_method(
            dir_path,
            extension,
            *args,
            base_filter_func=base_filter_func,
            **kwargs,
        )
        return file_names

    def check(self, *args, update_db=False, **kwargs):
        """Check for differences between the two directories.

        Arguments:
            *args: sent to the filter functions (should as minimum contain dir_path and extension).
            update_db: set to True if you want the check to also update the db.
        Additional keyword arguments:
            sent to the filter functions.

        """

        # PLEASE, REFACTOR ME!
        print(f"Comparing {self.from_dir} <=> {self.to_dir}")
        additional_filters = kwargs.pop("additional_filters", None)
        local_files = self._filter(
            self.local_connector,
            self.from_dir,
            *args,
            additional_filters=additional_filters,
            **kwargs,
        )
        external_files = list(
            self._filter(self.external_connector, self.to_dir, *args, **kwargs)
        )
        log.debug(external_files)
        number_of_local_files = 0
        number_of_external_duplicates = 0
        number_of_duplicates_out_of_sync = 0

        for f in local_files:
            number_of_local_files += 1

            self.make_external_name(f)
            external_name = self.external_name
            log.debug(f"{f.name} -> {self.external_name}")
            print(f" (*) {f.name}", end=" ")
            if external_name in external_files:
                code = 1
                print(f"[FOUND EXTERNAL]")
                number_of_external_duplicates += 1
                local_vals = self.checker.check(f)
                external_vals = self.checker.check(
                    external_name,
                    connector=self.external_connector,
                )

                same = True
                for k in local_vals:
                    print(f"     (L) {k}: {local_vals[k]}")
                    print(f"     (E) {k}: {external_vals[k]}")
                    if local_vals[k] != external_vals[k]:
                        same = False
                        number_of_duplicates_out_of_sync += 1
                print(f"     In sync: {same}")
            else:
                code = 0
                print("[ONLY LOCAL]")
            if update_db:
                print("     ! UPDATING DB")
                self.bookkeeper.register(f)
                if self.bookkeeper.code < 2:
                    self.bookkeeper.code = code
                if self.bookkeeper.code == 1:
                    self.bookkeeper.update_record(external_name, **local_vals)
            else:
                print()
        print()
        print(f" Total number of local files:    {number_of_local_files}")
        print(f" Files with external duplicates: {number_of_external_duplicates}")
        print(f" Files out of sync:              {number_of_duplicates_out_of_sync}")

    def run(self):
        """Copy the files that needs to be copied and update the db."""
        log.info("Running...")
        for f in self.file_names:
            del self.status
            self.make_external_name(f)
            log.debug(f"{f.name} -> {self.external_name}")
            self.bookkeeper.register(f)
            checks = self.checker.check(f)

            if self.bookkeeper.is_changed(**checks):
                self.status = ("changed", True)
                if self.mover_method(
                    f, self.external_name, connector=self.external_connector
                ):
                    self.status = ("moved", True)
                    self.bookkeeper.update_record(self.external_name, **checks)

            self.report()

    def report(self):
        status = self.status
        f1 = status["name"]
        f2 = status["external_name"]

        if status.get("changed", False):
            txt = f"[bold blue]{f1}"
            if status.get("moved", False):
                txt += f" -> {f2}[/]"
            else:
                txt += f"[/] -> [bold red blink]{f2} FAILED![/]"
        else:
            txt = f"[bold green]{f1} == {f2}[/]"

        log.info(txt, extra={"markup": True})

    def _create_external_name(self, f):
        return self.to_dir / f.name

    def make_external_name(self, f):
        name = self._create_external_name(f)
        self.external_name = name
        self.status = ("name", f.name)
        self.status = ("external_name", str(self.external_name))

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, pair: tuple):
        key, value = pair
        self._status[key] = value

    @status.deleter
    def status(self):
        self._status = {}

    @property
    def external_name(self):
        return self._external_name

    @external_name.setter
    def external_name(self, name):
        self._external_name = name


def simple_worker(
    base_directory_from=None,
    base_directory_to=None,
    db_name=None,
    dry_run=False,
):
    """Create a Worker for copying files locally.

    Args:
        base_directory_from: directory to copy from.
        base_directory_to: directory to copy to.
        db_name: name of the database.
        dry_run: set to True if you would like to run without updating or moving anything.

    Returns:
        simple worker that can copy files between two local folder.
    """
    db_name = db_name or os.environ["OELEO_DB_NAME"]
    base_directory_from = base_directory_from or Path(os.environ["OELEO_BASE_DIR_FROM"])
    base_directory_to = base_directory_to or Path(os.environ["OELEO_BASE_DIR_TO"])

    bookkeeper = SimpleDbHandler(db_name)
    checker = SimpleChecker()
    # mover = SimpleMover()

    log.info(
        f"[bold]from:[/] [bold green]{base_directory_from}[/]", extra={"markup": True}
    )
    log.info(
        f"[bold]to  :[/] [bold blue]{base_directory_to}[/]", extra={"markup": True}
    )

    worker = Worker(
        checker=checker,
        mover_method=simple_mover,
        from_dir=base_directory_from,
        to_dir=base_directory_to,
        bookkeeper=bookkeeper,
        dry_run=dry_run,
    )
    return worker


def ssh_worker(
    base_directory_from: Path | None = None,
    connector: SSHConnector | None = None,
    db_name: str | None = None,
    dry_run: bool = False,
):
    """Create a Worker with SSHConnector.

    Args:
        base_directory_from: directory to copy from.
        connector: an SSHConnector with external directory given.
        db_name: name of the database.
        dry_run: set to True if you would like to run without updating or moving anything.

    Returns:
        worker with SSHConnector attached to it.
    """
    if connector is None:
        raise ValueError("Required argument missing! connector cannot be None")

    db_name = db_name or os.environ["OELEO_DB_NAME"]
    base_directory_from = base_directory_from or Path(os.environ["OELEO_BASE_DIR_FROM"])
    base_directory_to = connector.directory

    bookkeeper = SimpleDbHandler(db_name)
    checker = ConnectedChecker()

    log.info(
        f"[bold]from:[/] [bold green]{base_directory_from}[/]", extra={"markup": True}
    )
    log.info(
        f"[bold]to  :[/] [bold blue]{connector.host}:{base_directory_to}[/]",
        extra={"markup": True},
    )

    worker = Worker(
        checker=checker,
        mover_method=connected_mover,
        from_dir=base_directory_from,
        to_dir=base_directory_to,
        external_connector=connector,
        bookkeeper=bookkeeper,
        dry_run=dry_run,
    )
    return worker



import getpass
import logging
import os
from pathlib import Path, PurePosixPath
from typing import Any, Protocol

import dotenv
from fabric import Connection

log = logging.getLogger("oeleo")

FabricRunResult = Any
Hash = str


class Connector(Protocol):
    """Connectors are used to establish a connection to the server and
    provide the functions and methods needed for the movers and checkers.

    Connectors are typically only needed for transactions between computers.
    If no connector is given to the oeleo worker, the worker will use the
    same methods for local and external files.
    """

    def connect(self, use_password: bool = False, **kwargs) -> None:
        ...

    def close(
        self,
    ):
        ...

    def base_filter_sub_method(self, glob_pattern: str = "*", **kwargs) -> list:
        ...

    def list_content(
        self, glob_pattern: str = "*", max_depth: int = 1, hide=False
    ) -> FabricRunResult:
        ...

    def calculate_checksum(self, f: Path, hide: bool = True) -> Hash:
        ...

    def move_func(self, *args, **kwargs) -> str:
        ...


class SSHConnector(Connector):
    def __init__(self, username=None, host=None, directory=None, is_posix=True):
        self.session_password = os.environ["OELEO_PASSWORD"]
        self.username = username or os.environ["OELEO_USERNAME"]
        self.host = host or os.environ["OELEO_EXTERNAL_HOST"]
        self.directory = directory or os.environ["OELEO_BASE_DIR_TO"]
        self.is_posix = is_posix
        self.c = None
        self._validate()

    def __str__(self):
        text = "SSHConnector"
        text += f"{self.username=}\n"
        text += f"{self.host=}\n"
        text += f"{self.directory=}\n"
        text += f"{self.is_posix=}\n"
        text += f"{self.c=}\n"

        return text

    def _validate(self):
        if self.is_posix:
            self.directory = PurePosixPath(self.directory)
        else:
            self.directory = Path(self.directory)

    def connect(self, use_password: bool = False, **kwargs) -> None:
        if use_password:
            connect_kwargs = {
                "password": os.environ["OELEO_PASSWORD"],
            }
        else:
            connect_kwargs = {
                "key_filename": [os.environ["OELEO_KEY_FILENAME"]],
            }
        self.c = Connection(
            host=self.host, user=self.username, connect_kwargs=connect_kwargs
        )

    def close(self):
        self.c.close()

    def __delete__(self, instance):
        if self.c is not None:
            self.c.close()

    def base_filter_sub_method(self, glob_pattern: str = "*", **kwargs: Any) -> list:
        log.info("base filter function for SSHConnector")
        log.info("got this glob pattern:")
        log.info(f"{glob_pattern}")

        if self.c is None:  # make this as a decorator ("@connected")
            log.debug("Connecting ...")
            self.connect()

        result = self.list_content(glob_pattern, hide=True)
        file_list = result.stdout.strip().split("\n")
        if self.is_posix:
            file_list = [PurePosixPath(f) for f in file_list]
        else:
            file_list = [Path(f) for f in file_list]  # OBS Linux -> Win not supported!
        return file_list

    def list_content(self, glob_pattern="*", max_depth=1, hide=False):

        if self.c is None:  # make this as a decorator ("@connected")
            log.debug("Connecting ...")
            self.connect()

        cmd = f"find {self.directory} -maxdepth {max_depth} -name '{glob_pattern}'"
        log.debug(cmd)
        result = self.c.run(cmd, hide=hide)
        if not result.ok:
            log.info("it failed - should raise an exception her (future work)")
        return result

    def calculate_checksum(self, f, hide=True):
        if self.c is None:  # make this as a decorator ("@connected")
            log.debug("Connecting ...")
            self.connect()

        cmd = f"md5sum {self.directory/f}"
        result = self.c.run(cmd, hide=hide)
        if not result.ok:
            log.info("it failed - should raise an exception her (future work)")
        checksum = result.stdout.strip().split()[0]
        return checksum

    def move_func(self, path: Path, to: Path, **kwargs):
        if self.c is None:  # make this as a decorator ("@connected")
            log.debug("Connecting ...")
            self.connect()

        try:
            log.debug(f"Copying {path} to {to}")
            result = self.c.put(str(path), remote=str(to))
        except Exception as e:
            print("GOT AN EXCEPTION DURING COPYING FILE")
            print(f"FROM     : {path}")
            print(f"TO       : {to}")
            print(f"EXCEPTION:")
            print(e)
            return False
        return True


def register_password(pwd: str = None) -> None:
    print(" Register password ".center(80, "="))
    if pwd is None:
        session_password = getpass.getpass(prompt="Password: ")
        os.environ["OELEO_PASSWORD"] = session_password
    print(" Done ".center(80, "="))


def main():
    dotenv.load_dotenv()
    local_dir = Path(r"C:\scripting\processing_cellpy\raw")

    external_dir = PurePosixPath("/home/jepe@ad.ife.no/Temp")
    external_host = os.environ["EXTERNAL_TEST_HOST"]
    username = os.environ["OELEO_USERNAME"]
    keyname = os.environ["OELEO_KEY_FILENAME"]
    password = os.environ["OELEO_PASSWORD"]

    register_password()
    session_password = os.environ["OELEO_PASSWORD"]
    c = Connection(
        host=external_host,
        user=username,
        connect_kwargs={
            "password": session_password,
            # "key_filename": [keyname],
        },
    )

    result = c.run(f"ls {external_dir}")
    print(result)


if __name__ == "__main__":
    main()



import random
import shutil
from pathlib import Path


def mock_mover(path: Path, to: Path, *args, **kwargs):
    print(f"COPYING {path} -> {to}")
    success = random.choice([True, False])
    print(f"success={success}")
    return success


def simple_mover(path: Path, to: Path, *args, **kwargs):
    try:
        shutil.copyfile(path, to)
        return True
    except OSError:
        print("Could not copy this file - destination most likely not writable!")
        return False


def connected_mover(path: Path, to: Path, connector=None, *args, **kwargs):
    """Copies files using the method implemented in the connector."""
    if connector is None:
        move_func = simple_mover
    else:
        move_func = connector.move_func
    try:
        success = move_func(path, to)
        return success
    except OSError:
        print("Could not copy this file - destination most likely not writable!")
        return False
